/*
	Copyright  1995-2001, The AROS Development Team. All rights reserved.
	$Id: listviewgclass.c,v 1.1.1.1 2005/03/15 15:54:23 laire Exp $
 
	Desc: Internal GadTools listview class.
	Lang: English
*/

#undef AROS_ALMOST_COMPATIBLE
#define AROS_ALMOST_COMPATIBLE 1

#include <proto/exec.h>
#include <exec/libraries.h>
#include <exec/memory.h>
#include <proto/dos.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/icclass.h>
#include <intuition/intuition.h>
#include <utility/tagitem.h>
#include <proto/alib.h>
#include <proto/utility.h>
#include <proto/gadtools.h>

#include <string.h> /* memset() */

#if 0
#define SDEBUG 1
#define DEBUG 1
#endif

#include "gadtools_intern.h"

/**********************************************************************************************/

#define G(x) ((struct Gadget *)(x))
#define EG(x) ((struct ExtGadget *)(x))

#define GadToolsBase ((struct GadToolsBase_intern *)cl->cl_UserData)

/**********************************************************************************************/

/* Spacing between scroller and listview */
#define SCROLLER_SPACING 2

/****************************************************************************************/


struct LVGData
{
	Object		*listview;
	Object		*scroller;
	WORD         scrollerx;
	WORD         scrollerw;
	BOOL         scrollerdown;
};

/**********************************************************************************************/

STATIC IPTR listviewg_new(Class *cl, Object *o, struct opSet *msg)
{
	Object *lvgad = NULL;
	struct Gadget *showselgad = LV_SHOWSELECTED_NONE;
	Object *scrollergad = NULL;
	struct TagItem *taglist, *tag;
	int left = 0, top = 0, width = 0, height = 0;
	struct DrawInfo *dri = NULL;
	struct TextAttr *tattr;

#if CORRECT_LISTVIEWHEIGHT
	WORD viewheight;
#endif
	WORD ysize, totalitemheight = 0, scrollerwidth = 16;

	static const struct TagItem lv_map[] =
	{
	    {GTLV_Top,	 PGA_Top},
	    {TAG_DONE			}
	};

	static const struct TagItem sc_map[] =
	{
	    {PGA_Top,	 GTLV_Top},
	    {TAG_DONE			}
	};

	/* listviewg tags */

	struct TagItem lvgtags[] =
	{
	    {GA_FollowMouse		 , FALSE}, /* 0 */
	    //{GA_Bounds		, (IPTR)&bbox		}, /*   */
	    {TAG_MORE		 , (IPTR)NULL}  /* 1 */
	};

	/* listview tags */

	struct TagItem lvtags[] =
	{
	    {GA_Width		 , 0L}, /* 0  */
	    {GA_Height		, 0L			}, /* 1  */
	    {GA_DrawInfo		, (IPTR)NULL		}, /* 2  */
	    {GTLV_Top		, 0L			}, /* 3  */
	    {GTLV_MakeVisible	, 0L			}, /* 4  */
	    {GTLV_Labels		, (IPTR)NULL		}, /* 5  */
	    {GTLV_Selected		, ~0L			}, /* 6  */
	    {GTLV_ItemHeight	, 0L			}, /* 7  */
	    {GTLV_CallBack		, (IPTR)NULL		}, /* 8  */
	    {GTLV_MaxPen		, 0L			}, /* 9  */
	    {GTLV_ReadOnly		, 0L			}, /* 10 */
	    {LAYOUTA_Spacing	, 0L			}, /* 11 */
	    {GA_TextAttr		, (IPTR)NULL		}, /* 12 */
	    {GTLV_ShowSelected	, (IPTR)showselgad	}, /* 13 */
	    {ICA_MAP		, (IPTR)&lv_map		}, /* 14 */
	    {ICA_TARGET		, (IPTR)NULL		}, /* 15 */
	    {TAG_DONE					}
	};

	/* scroller tags */

	struct TagItem scrtags[] =
	{
	    {GA_Left	 , 0L}, /* 0 */
	    {GA_Top			, 0L		}, /* 1 */
	    {GA_Width		, 0L		}, /* 2 */
	    {GA_Height		, 0L		}, /* 3 */
	    {GA_DrawInfo		, (IPTR)NULL	}, /* 4 */
	    {GTSC_Arrows		, 16		}, /* 5 */
	    {PGA_Freedom		, LORIENT_VERT	}, /* 6 */
	    {ICA_MAP		, (IPTR)&sc_map	}, /* 7 */
	    {TAG_DONE				}
	    /* The listview will initialize the scrollers top, visible & total */
	};

	struct TagItem settags[] =
	{
	    {ICA_TARGET		 , (IPTR)NULL},
	    {TAG_DONE				}
	};

	struct opSet newmsg;

	EnterFunc(bug("Listviewg::New()\n"));

	/* Could use GetTagData(), but this is faster */
	taglist = msg->ops_AttrList;
	while ((tag = NextTagItem(&taglist)))
	{
		IPTR tidata = tag->ti_Data;

		switch (tag->ti_Tag)
		{
		case GA_Left:
			left = (WORD)tidata;
			break;
		case GA_Top:
			top = (WORD)tidata;
			break;
		case GA_Width:
			width = (WORD)tidata;
			break;
		case GA_Height:
			height = (WORD)tidata;
			break;

		case GA_DrawInfo:
			dri = (APTR)tidata;
			lvtags[2].ti_Data = tidata;
			scrtags[4].ti_Data = tidata;
			break;

		case GTLV_Top:
			lvtags[3].ti_Data = tidata;
			break;
		case GTLV_MakeVisible:
			lvtags[4].ti_Data = tidata;
			break;
		case GTLV_Labels:
			lvtags[5].ti_Data = tidata;
			break;
		case GTLV_Selected:
			lvtags[6].ti_Data = tidata;
			break;
		case GTLV_ItemHeight:
			lvtags[7].ti_Data = tidata;
			break;
		case GTLV_CallBack:
			lvtags[8].ti_Data = tidata;
			break;
		case GTLV_MaxPen:
			lvtags[9].ti_Data = tidata;
			break;
		case GTLV_ReadOnly:
			lvtags[10].ti_Data = tidata;
			break;
		case LAYOUTA_Spacing:
			lvtags[11].ti_Data = tidata;
			break;
		case GA_TextAttr:
			lvtags[12].ti_Data = tidata;
			break;

		case GTLV_ShowSelected:
			showselgad  = (struct Gadget *)tidata;
			lvtags[13].ti_Data = (IPTR)showselgad;
			break;

		case GTLV_ScrollWidth:
			scrollerwidth = (UWORD)tidata;
		}

	} /* while (iterate taglist) */

	if (!dri)
	{
		ReturnPtr ("listviewg_new", IPTR, 0);
	}

	/* Callback supplied ? */
	if (!lvtags[8].ti_Data)
		lvtags[8].ti_Tag = TAG_IGNORE; /* Don't pass GTLV_Callback, NULL */

	tattr = (APTR)lvtags[12].ti_Data;

	if (tattr) /* Text Attr supplied ? */
	{
		struct TextFont *font;

		font = OpenFont(tattr);
		if (font)
		{
			ysize = font->tf_YSize;
			CloseFont(font);
		}
		else
		{
			ysize = tattr->ta_YSize;
		}
	}
	else
	{
		ysize = dri->dri_Font->tf_YSize;
	}

	/* If not set allready, set ItemHeight */
	if (lvtags[7].ti_Data == 0)
		lvtags[7].ti_Data = ysize;

	/* If not set allready, set Spacing */
	if (lvtags[11].ti_Data == 0)
		lvtags[11].ti_Data = LV_DEF_INTERNAL_SPACING;

	D(bug("listviewg_new: showselgad %p\n", showselgad));

	/* Set the listview height according to showsel gadget height */
	if ((showselgad) && (showselgad != LV_SHOWSELECTED_NONE))
	{
		/* The showselected string gadget must have the same width as the
		   listview gadget, otherwise fail (AmigaOS does the same). Fail
		   also if the showselected string gadget is too high*/

		D(bug("listviewg_new: width %d height %d strw %d (%d) strh %d (%d)\n",
		      width, height, EG(showselgad)->BoundsWidth, EG(showselgad)->Width,
		      EG(showselgad)->BoundsHeight, EG(showselgad)->Height));

		if ((EG(showselgad)->BoundsWidth != width) ||
		        (EG(showselgad)->BoundsHeight >= height))
		{
			ReturnPtr ("listviewg_new", IPTR, 0);
		}

		/* the showselected string gadget shrinks the height of the listview */

		height -= EG(showselgad)->BoundsHeight;

		/* the showselected string gadget will get its position automatically
		   fixed to be under the listview gadget */

		EG(showselgad)->BoundsLeftEdge = left;
		EG(showselgad)->BoundsTopEdge  = top + height;
		EG(showselgad)->LeftEdge       = EG(showselgad)->BoundsLeftEdge + BORDERSTRINGSPACINGX;
		EG(showselgad)->TopEdge        = EG(showselgad)->BoundsTopEdge  + BORDERSTRINGSPACINGY;

		D(bug("listviewg_new: Showselected gadget specified"));

	}

	/* Set the listview width according to scroller width + some spacing */
	width -= scrollerwidth + SCROLLER_SPACING;
	lvtags[0].ti_Data = width;

	/* GTLV_ItemHeight + LAYOUTA_Spacing */
	totalitemheight = lvtags[7].ti_Data + lvtags[11].ti_Data;

#if CORRECT_LISTVIEWHEIGHT
	/* stegerg: I think it looks better without this adjustment. Think of
	   GTLV_ShowSelected and aligning with other gadgets */

	/* Adjust listview height so that an exact number of items fits in it */
	viewheight = height - (2 * LV_BORDER_Y);
	height -= (viewheight % totalitemheight);

#endif

	lvtags[1].ti_Data = height;

	/* Reinsert the modified dimension attrs into the listview taglist */

	D(bug("listviewg_new: Dimension passed to listview: w=%d, h=%d\n", width, height));

	D(bug("listviewg_new: scrollerwidth %ld\n",scrollerwidth));

	/* Create a scroller object to use with the listviev */
	scrtags[0].ti_Data = width + SCROLLER_SPACING;
	scrtags[2].ti_Data = scrollerwidth;
	scrtags[3].ti_Data = height;

	scrollergad = NewObjectA(GadToolsBase->scrollergclass, NULL, scrtags);

	D(bug("listviewg_new: Scroller gadget created: %p\n", scrollergad));

	if (!scrollergad)
	{
		ReturnPtr ("listviewg_new", IPTR, 0);
	}


	lvtags[15].ti_Data = (IPTR)scrollergad;

	lvgad = NewObjectA(GadToolsBase->listviewclass, NULL, lvtags);

	D(bug("listviewg_new: Listview gadget created: %p\n", lvgad));

	if (!lvgad)
		goto failure;

	settags[0].ti_Data = (IPTR)lvgad;
	SetAttrsA(scrollergad, settags);

	lvgtags[1].ti_Data = (IPTR)msg->ops_AttrList;

	newmsg.MethodID = OM_NEW;
	newmsg.ops_AttrList = lvgtags;
	newmsg.ops_GInfo = NULL;

	o = (Object *)DoSuperMethodA(cl, o, (Msg)&newmsg);
	if (o)
	{
		struct LVGData *data = INST_DATA(cl, o);
		struct opMember addm;

		DEBUG_CREATESCROLLER(bug("listviewg::New: listviewg %p\n", o));

		data->listview = lvgad;
		data->scroller = scrollergad;
		data->scrollerx = scrtags[0].ti_Data;
		data->scrollerw = scrtags[2].ti_Data;

		addm.MethodID = OM_ADDMEMBER;
		addm.opam_Object = lvgad;
		DoMethodA(o, (Msg)&addm);

		addm.MethodID = OM_ADDMEMBER;
		addm.opam_Object = scrollergad;
		DoMethodA(o, (Msg)&addm);

		ReturnPtr ("listviewg::New", IPTR, (IPTR)o);
	}

failure:
	DisposeObject(scrollergad);
	DisposeObject(lvgad);

	ReturnPtr ("listviewg::New", IPTR, 0);
}

/**********************************************************************************************/

STATIC IPTR listviewg_set(Class * cl, Object * o, struct opUpdate * msg)
{
	IPTR		retval = 0UL;
	struct TagItem	*tag, *tstate;
	struct TagItem	lvtags[5] =
	{
	    {TAG_IGNORE, 0},
		{TAG_IGNORE, 0},
		{TAG_IGNORE, 0},
		{TAG_IGNORE, 0},
		{TAG_DONE}
	};
	BOOL has_lvtag = FALSE, has_non_lvtag = FALSE;

	D(bug("listviewg_set(cl 0x%lx o 0x%lx msg 0x%lx)\n",cl,o,msg));

	tstate = msg->opu_AttrList;

	while ((tag = NextTagItem(&tstate)))
	{
		D(bug("listviewg_set: tag 0x%08lx 0x%08lx\n", tag->ti_Tag, tag->ti_Data));

		switch (tag->ti_Tag)
		{
		case GTLV_Top:
			lvtags[0] = *tag;
			has_lvtag = TRUE;
			break;
		case GTLV_MakeVisible:
			lvtags[1] = *tag;
			has_lvtag = TRUE;
			break;
		case GTLV_Labels:
			lvtags[2] = *tag;
			has_lvtag = TRUE;
			break;
		case GTLV_Selected:
			lvtags[3] = *tag;
			has_lvtag = TRUE;
			break;

		default:
			has_non_lvtag = TRUE;
			break;
		}

	}

	if (has_lvtag)
	{
		struct opUpdate method;
		struct LVGData *data = INST_DATA(cl, o);

		D(bug("listviewg_set: set scroller tags\n"));

		method.MethodID = msg->MethodID;
		method.opu_AttrList = lvtags;
		method.opu_GInfo = msg->opu_GInfo;
		method.opu_Flags = msg->opu_Flags; /* only valid for OM_UPDATE. Oh well. */

		retval = DoMethodA(data->listview, (Msg)&method);
	}

	if (has_non_lvtag)
	{
		D(bug("listviewg_set: set superclass tags\n"));

		retval += DoSuperMethodA(cl, o, (Msg)msg);
	}

	if ((retval) && ((msg->MethodID != OM_UPDATE) || (cl == OCLASS(o))))
	{
		struct RastPort *rp;

		D(bug("listviewg_set: redrawing list\n"));

		rp = ObtainGIRPort(msg->opu_GInfo);
		if (rp)
		{
			DoMethod(o, GM_RENDER, (IPTR)msg->opu_GInfo, (IPTR)rp, GREDRAW_REDRAW);
			ReleaseGIRPort(rp);
			retval = FALSE;
		}
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR listviewg_get(Class * cl, Object * o, struct opGet *msg)
{
	IPTR		retval;

	switch(msg->opg_AttrID)
	{
	case GTA_GadgetKind:
	case GTA_ChildGadgetKind:
		*msg->opg_Storage = LISTVIEW_KIND;
		retval = 1UL;
		break;

	case GTLV_Top:
	case GTLV_Labels:
	case GTLV_Selected:
		{
			struct LVGData *data = INST_DATA(cl, o);
			retval = DoMethodA(data->listview, (Msg)msg);
		}
		break;

	default:
		retval = DoSuperMethodA(cl, o, (Msg)msg);
		break;
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR listviewg_gadgetidcmpmsg(Class *cl, Object *o, struct gtpGadgetIDCMPMsg *msg)
{
	struct LVGData *data = INST_DATA(cl, o);
	IPTR val;

	switch (msg->IntuiMsg->Class)
	{
	case IDCMP_GADGETDOWN:
		return TRUE;

	case IDCMP_GADGETUP:
		//jDc: need to determine here if we click the scroller or list
		{
			if ((msg->IntuiMsg->MouseX >= ((struct Gadget *)(o))->LeftEdge + data->scrollerx) && (msg->IntuiMsg->MouseX < ((struct Gadget *)(o))->LeftEdge + data->scrollerx + data->scrollerw)) return TRUE;

			GetAttr(GTLV_Selected, data->listview, &val);
			if ((UWORD)val == 0xffff)
				return TRUE;
			msg->IntuiMsg->Code = (UWORD)val;
		}
		break;

	case IDCMP_RAWKEY:
	case IDCMP_VANILLAKEY:
		/* wheel events ONLY */
		{
			ULONG total,top;
			GetAttr(GTLV_Top,o,&top);

			switch (msg->IntuiMsg->Code)
			{
				case NM_WHEEL_UP:
					top -= (top > 2) ? 3 : top;
					SetGadgetAttrs(o,msg->IntuiMsg->IDCMPWindow,0,GTLV_Top,top,TAG_DONE);
					break;

				case NM_WHEEL_DOWN:
					GetAttr(GTA_Listview_NumEntries,data->listview,&total);
					top = (top + 3 < total) ? (top + 3) : (total - 1);
					SetGadgetAttrs(o,msg->IntuiMsg->IDCMPWindow,0,GTLV_Top,top,TAG_DONE);
					break;
			}
		}
		return TRUE;
		break;

	case IDCMP_INTUITICKS:
	case IDCMP_MOUSEMOVE:
		return TRUE;
		break;
	}

	return FALSE;
}

/**********************************************************************************************/

AROS_UFH3S(IPTR, dispatch_listviewgclass,
		   AROS_UFHA(Class *,  cl,  A0),
		   AROS_UFHA(Object *, o,   A2),
		   AROS_UFHA(Msg,      msg, A1)
		  )
{
	AROS_USERFUNC_INIT

	IPTR retval;

	D(bug("dispatch_listviewgclass: cl 0x%lx o 0x%lx msg 0x%lx method %ld (0x%lx)\n",
	      cl,o,msg,msg->MethodID,msg->MethodID));

	switch(msg->MethodID)
	{
	case OM_NEW:
		D(bug("dispatch_listviewgclass: OM_NEW\n"));
		retval = listviewg_new(cl, o, (struct opSet *)msg);
		break;

	case OM_UPDATE:
	case OM_SET:
		D(bug("dispatch_listviewgclass: %s\n",
		      msg->MethodID == OM_SET ? "OM_SET" : "OM_UPDATE"));
		retval = DoSuperMethodA(cl, o, msg);
		retval += (IPTR)listviewg_set(cl, o, (struct opUpdate *)msg);
		break;

	case OM_GET:
		D(bug("dispatch_listviewgclass: OM_GET\n"));
		retval = (IPTR)listviewg_get(cl, o, (struct opGet *)msg);
		break;

	case GTM_GadgetIDCMPMsg:
		D(bug("dispatch_scrollergclass: GTM_GadgetIDCMPMsg\n"));
		retval = listviewg_gadgetidcmpmsg(cl, o, (struct gtpGadgetIDCMPMsg *)msg);
		break;

	default:
		retval = DoSuperMethodA(cl, o, msg);
		break;
	} /* switch */

	D(bug("dispatch_listviewgclass: retval 0x%lx\n",retval));

	return (retval);

	AROS_USERFUNC_EXIT
}  /* dispatch_Gtlvclass */

/**********************************************************************************************/

#undef GadToolsBase

Class *makelistviewgclass(struct GadToolsBase_intern * GadToolsBase)
{
	Class *cl;

	ObtainSemaphore(&GadToolsBase->classsema);

	cl = GadToolsBase->listviewgclass;
	if (!cl)
	{
		if (makegroupclass(GadToolsBase) && makelistviewclass(GadToolsBase) &&
		        makescrollergclass(GadToolsBase))
		{
			cl = MakeClass(NULL, NULL, GadToolsBase->groupclass, sizeof(struct LVGData), 0UL);
			if (cl)
			{
				cl->cl_Dispatcher.h_Entry = (APTR) AROS_ASMSYMNAME(dispatch_listviewgclass);
				cl->cl_Dispatcher.h_SubEntry = NULL;
				cl->cl_UserData = (IPTR) GadToolsBase;

				GadToolsBase->listviewgclass = cl;
			}
		}
	}

	ReleaseSemaphore(&GadToolsBase->classsema);

	return cl;
}

/**********************************************************************************************/

VOID freelistviewgclass(Class *cl, struct GadToolsBase_intern *GadToolsBase)
{
	FreeClass(cl);
}

/**********************************************************************************************/
