/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: scrollerclass.c,v 1.1.1.1 2005/03/15 15:54:23 laire Exp $
 
    Desc: Internal GadTools scroller class.
    Lang: English
*/

#undef AROS_ALMOST_COMPATIBLE
#define AROS_ALMOST_COMPATIBLE 1

#include <proto/exec.h>
#include <exec/libraries.h>
#include <exec/memory.h>
#include <proto/dos.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/intuition.h>
#include <intuition/cghooks.h>
#include <graphics/rastport.h>
#include <graphics/text.h>
#include <utility/tagitem.h>
#include <devices/inputevent.h>
#include <proto/alib.h>
#include <proto/utility.h>

#include <string.h> /* memset() */

#if 0
#define SDEBUG 1
#define DEBUG 1
#endif

#include "gadtools_intern.h"

/**********************************************************************************************/

#define G(x) ((struct Gadget *)(x))
#define EG(X) ((struct ExtGadget *)(x))

#define GadToolsBase ((struct GadToolsBase_intern *)cl->cl_UserData)

/**********************************************************************************************/

struct ScrollerData
{
	Object		*frame;
};

/**********************************************************************************************/

#define opU(x) ((struct opUpdate *)msg)

STATIC IPTR scroller_set(Class * cl, Object * o, struct opSet * msg)
{
	IPTR		retval = 0UL;
	struct TagItem	*tag, *tstate, tags[] =
	    {
		    {
			    PGA_Total	, 0
		    },
		    {PGA_Top	, 0	},
		    {PGA_Visible	, 0	},
		    {TAG_MORE		}
	    };

	//struct ScrollerData *data = INST_DATA(cl, o);

	D(bug("scroller_set(cl 0x%lx o 0x%lx msg 0x%lx)\n",cl,o,msg));

	tags[3].ti_Data = (IPTR)msg->ops_AttrList;

	/* Get old values */
	DoSuperMethod(cl, o, OM_GET, PGA_Total,	(ULONG) &(tags[0].ti_Data));
	DoSuperMethod(cl, o, OM_GET, PGA_Top,	(ULONG) &(tags[1].ti_Data));
	DoSuperMethod(cl, o, OM_GET, PGA_Visible,	(ULONG) &(tags[2].ti_Data));

	D(bug("scroller_set: Old Total %ld Top %ld Visible %ld\n",
	      tags[0].ti_Data,
	      tags[1].ti_Data,
	      tags[2].ti_Data));

	tstate = msg->ops_AttrList;

	while ((tag = NextTagItem((struct TagItem **)&tstate)))
	{

		switch (tag->ti_Tag)
		{
		case GTSC_Total:
			D(bug("scroller_set: GTSC_Total %ld\n",tag->ti_Data));
			tags[0].ti_Data  = tag->ti_Data;
			break;

		case GTSC_Top:
			D(bug("scroller_set: GTSC_Top %ld\n",tag->ti_Data));
			tags[1].ti_Data  = tag->ti_Data;
			break;

		case GTSC_Visible:
			D(bug("scroller_set: GTSC_Visible %ld\n",tag->ti_Data));
			tags[2].ti_Data  = tag->ti_Data;
			break;

		case GTA_Scroller_Dec:
			D(bug("scroller_set: GTA_Scroller_Dec\n"));
			if (tags[1].ti_Data > 0)
			{
				((ULONG)tags[1].ti_Data) --;
				retval = 1UL;
			}
			break;

		case GTA_Scroller_Inc:
			/* Top < (Total - Visible) ? */
			D(bug("scroller_set: GTA_Scroller_Inc\n"));
			if (tags[1].ti_Data < (tags[0].ti_Data - tags[2].ti_Data))
			{
				((ULONG)tags[1].ti_Data) ++;
				retval = 1UL;
			}
			break;

		} /* switch (tag->ti_Tag) */

	} /* while ((tag = NextTagItem((const struct TagItem **)&tstate))) */

	DoSuperMethod(cl, o, OM_SET, (ULONG) tags, (ULONG) msg->ops_GInfo);

	return (retval);
}

/**********************************************************************************************/

STATIC IPTR scroller_get(Class * cl, Object * o, struct opGet *msg)
{
	//struct ScrollerData *data = INST_DATA(cl, o);
	struct opGet	cloned_msg = *msg;
	IPTR		retval = 1UL;

	D(bug("scroller_get: AttrID 0x%lx\n",msg->opg_AttrID));
	switch (msg->opg_AttrID)
	{
	case GTSC_Top:
		D(bug("scroller_get: GTSC_Top\n"));
		cloned_msg.opg_AttrID = PGA_Top;
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;

	case GTSC_Total:
		D(bug("scroller_get: GTSC_Total\n"));
		cloned_msg.opg_AttrID = PGA_Total;
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;

	case GTSC_Visible:
		D(bug("scroller_get: GTSC_Visible\n"));
		cloned_msg.opg_AttrID = PGA_Visible;
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;

	default:
		retval = DoSuperMethodA(cl, o, (Msg)msg);
		break;
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR scroller_new(Class * cl, Object * o, struct opSet *msg)
{
	struct ScrollerData *data;
	struct DrawInfo	*dri;
	Object		*frame;
	struct TagItem	fitags[] =
	    {
		    {
			    IA_Width	, 0UL
		    },
		    {IA_Height	, 0UL		},
		    {IA_Resolution	, 0UL		},
		    {IA_FrameType	, FRAME_BUTTON	},
		    {IA_EdgesOnly	, TRUE		},
		    {TAG_DONE			}
	    };

	EnterFunc(bug("Scroller::New()\n"));

	dri = (struct DrawInfo *)GetTagData(GA_DrawInfo, NULL, msg->ops_AttrList);

	fitags[0].ti_Data = GetTagData(GA_Width, 0, msg->ops_AttrList) + BORDERPROPSPACINGX * 2;
	fitags[1].ti_Data = GetTagData(GA_Height, 0, msg->ops_AttrList) + BORDERPROPSPACINGY * 2;
	fitags[2].ti_Data = (dri->dri_Resolution.X << 16) + dri->dri_Resolution.Y;

	frame = NewObjectA(NULL, FRAMEICLASS, fitags);

	D(bug("Scroller::New: frame %p\n", frame));

	if (frame)
	{
		o = (Object *)DoSuperMethodA(cl, o, (Msg)msg);

		D(bug("Scroller::New: scroller %p\n", o));

		if (o)
		{
			data = INST_DATA(cl, o);
			data->frame = frame;

			scroller_set(cl, o, msg);
		}
	}
	else
	{
		o = NULL;
	}

	ReturnPtr("Scroller::New", IPTR, (IPTR)o);
}

/**********************************************************************************************/

STATIC IPTR scroller_render(Class *cl, Object *o, struct gpRender *msg)
{
	IPTR		retval;

	if (msg->gpr_Redraw == GREDRAW_REDRAW)
	{
		struct ScrollerData *data = INST_DATA(cl, o);

		DrawImageState(msg->gpr_RPort,
		               (struct Image *)data->frame,
		               G(o)->LeftEdge - BORDERPROPSPACINGX,
		               G(o)->TopEdge  - BORDERPROPSPACINGY,
		               IDS_NORMAL,
		               msg->gpr_GInfo->gi_DrInfo);
	}

	retval = DoSuperMethodA(cl, o, (Msg)msg);

	ReturnInt("Scroller::Render", IPTR, retval);
}

/**********************************************************************************************/

STATIC IPTR scroller_dispose(Class *cl, Object *o, Msg msg)
{
	struct ScrollerData *data = INST_DATA(cl, o);

	if (data->frame) DisposeObject(data->frame);

	return DoSuperMethodA(cl, o, msg);
}

/**********************************************************************************************/

AROS_UFH3S(IPTR, dispatch_scrollerclass,
           AROS_UFHA(Class *, cl, A0),
           AROS_UFHA(Object *, o, A2),
           AROS_UFHA(Msg, msg, A1)
          )
{
	AROS_USERFUNC_INIT

	IPTR retval;

	D(bug("dispatch_scrollerclass: cl 0x%lx o 0x%lx msg 0x%lx\n",cl,o,msg));
	switch (msg->MethodID)
	{
	case OM_NEW:
		D(bug("dispatch_scrollerclass: OM_NEW\n"));
		retval = scroller_new(cl, o, (struct opSet *) msg);
		break;

	case OM_DISPOSE:
		D(bug("dispatch_scrollerclass: OM_DISPOSE\n"));
		retval = scroller_dispose(cl, o, msg);
		break;

	case OM_UPDATE:
		D(bug("dispatch_scrollerclass: OM_UPDATE\n"));
	case OM_SET:
		D(bug("dispatch_scrollerclass: OM_SET\n"));
		retval = scroller_set(cl, o, (struct opSet *) msg);
		/* If we have been subclassed, OM_UPDATE should not cause a GM_RENDER
		 * because it would circumvent the subclass from fully overriding it.
		 * The check of cl == OCLASS(o) should fail if we have been
		 * subclassed, and we have gotten here via DoSuperMethodA().
		 */
		if ( retval && ( (msg->MethodID != OM_UPDATE) ||  (cl == OCLASS(o)) ) )
		{
			struct GadgetInfo *gi = ((struct opSet *)msg)->ops_GInfo;
			if (gi)
			{
				struct RastPort *rp = ObtainGIRPort(gi);
				if (rp)
				{
					DoMethod(o, GM_RENDER, (ULONG) gi, (ULONG) rp, GREDRAW_REDRAW);
					ReleaseGIRPort(rp);
				} /* if */
			} /* if */
		} /* if */
		break;

	case OM_GET:
		D(bug("dispatch_scrollerclass: OM_GET\n"));
		retval = scroller_get(cl, o, (struct opGet *)msg);
		break;

	case GM_RENDER:
		D(bug("dispatch_scrollerclass: OM_RENDER\n"));
		retval = scroller_render(cl, o, (struct gpRender *)msg);
		break;

	default:
		D(bug("dispatch_scrollerclass: gate..\n"));
		retval = DoSuperMethodA(cl, o, msg);
		break;
	}

	D(bug("dispatch_scrollerclass: retval 0x%lx\n",retval));
	return retval;

	AROS_USERFUNC_EXIT
}

/**********************************************************************************************/

#undef GadToolsBase

Class *makescrollerclass(struct GadToolsBase_intern * GadToolsBase)
{
	Class *cl;

	ObtainSemaphore(&GadToolsBase->classsema);

	cl = GadToolsBase->scrollerclass;
	if (!cl)
	{
		cl = MakeClass(NULL, PROPGCLASS, NULL, sizeof(struct ScrollerData), 0UL);
		if (cl)
		{
			cl->cl_Dispatcher.h_Entry = (APTR) AROS_ASMSYMNAME(dispatch_scrollerclass);
			cl->cl_Dispatcher.h_SubEntry = NULL;
			cl->cl_UserData = (IPTR) GadToolsBase;

			GadToolsBase->scrollerclass = cl;
		}
	}

	ReleaseSemaphore(&GadToolsBase->classsema);

	return cl;
}

/**********************************************************************************************/
