/*
	Copyright  1995-2001, The AROS Development Team. All rights reserved.
	$Id: slidergclass.c,v 1.1.1.1 2005/03/15 15:54:23 laire Exp $
 
	Desc: Internal GadTools slider class.
	Lang: English
*/

#undef AROS_ALMOST_COMPATIBLE
#define AROS_ALMOST_COMPATIBLE 1

#include <exec/libraries.h>
#include <exec/memory.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/icclass.h>
#include <intuition/intuition.h>
#include <utility/tagitem.h>
#include <proto/alib.h>
#include <proto/utility.h>
#include <proto/gadtools.h>

#include <string.h> /* memset() */

#if 0
#define SDEBUG 1
#define DEBUG 1
#endif

#include "gadtools_intern.h"

#undef DEBUG_CREATESLIDER
#define DEBUG_CREATESLIDER(x)	;

/**********************************************************************************************/

#define G(x) ((struct Gadget *)(x))
#define EG(X) ((struct ExtGadget *)(x))

#define GadToolsBase ((struct GadToolsBase_intern *)cl->cl_UserData)

/**********************************************************************************************/

AROS_UFP2 (void, puttostr,
		   AROS_UFPA(UBYTE, chr, D0),
		   AROS_UFPA(STRPTR *,strPtrPtr,A3)
		  );

/**********************************************************************************************/

struct SGData
{
	Object		*slider;
	struct TextFont	*font;
	UBYTE		justification;
	UWORD		code;
	BOOL		immediate;
	BOOL		relverify;
	APTR		levelformat;
	APTR		dispfunc;
	WORD		level_x;
	WORD		level_y;
	WORD		level_width;
	WORD		level_height;
	WORD		erase_x1;
	WORD		erase_y1;
	WORD		erase_x2;
	WORD		erase_y2;
	BOOL		draw_level;
	WORD		last_level;
};

/**********************************************************************************************/

STATIC IPTR sliderg_new(Class *cl, Object *o, struct opSet *msg)
{
	struct TagItem *tag, *taglist;
	BOOL immediate = FALSE, relverify = FALSE;

	/* Tags for the sliderg */

	struct TagItem sgtags[] =
	    {
		    {
			    GA_FollowMouse	, TRUE
		    },
		    {TAG_MORE	, (IPTR)NULL	}
	    };

	/* Tags for the slider */

	struct TagItem stags[] =
	    {
		    {
			    GA_Left	, BORDERPROPSPACINGX	/* 0 */
		    },
		    {GA_Top		, BORDERPROPSPACINGY},	/* 1 */
		    {GA_Width	, 0		},		/* 2 */
		    {GA_Height	, 0		},		/* 3 */
		    {GA_DrawInfo	, (IPTR)NULL	},	/* 4 */
		    {GA_UserData	, 0	},		/* 5 */
		    {GTSL_Min	, 0		},		/* 6 */
		    {GTSL_Max	, 15		},		/* 7 */
		    {GTSL_Level	, 0		},		/* 8 */
		    {PGA_Freedom	, FREEHORIZ	},	/* 9 */
		    {PGA_Borderless	, TRUE		},	/* 10 */
#ifdef __MORPHOS__
		    {PGA_NewLook    , FALSE         },		/* 11 */
#else
		    {PGA_NewLook    , TRUE          },		/* 11 */
#endif
		    {PGA_NotifyBehaviour, PG_BEHAVIOUR_NICE },	/* 12 */
		    {PGA_RenderBehaviour, PG_BEHAVIOUR_NICE },	/* 13 */
		    //{ICA_MAP	, s_map		},
		    {TAG_DONE			}
	    };

	struct TextAttr *tattr = NULL;
	//struct TextFont *font = NULL;
	struct DrawInfo *dri = NULL;
	STRPTR lformat = NULL;
	WORD lmaxlen = 0;
	LONG lmaxpixellen = 0L;
	UBYTE lplace = PLACETEXT_LEFT;
	UBYTE justification = GTJ_LEFT;
	WORD level = 0;
	APTR dispfunc = NULL;
	WORD x, y;
	UWORD ysize = 8;
	WORD width, height;

	Object *slidergad = NULL;
	struct opSet newmsg;


	DEBUG_CREATESLIDER(bug("sliderg::New:\n"));

	/* Prevent gcc warning, width/heigth tags *are* there */
	width = 0;
	height = 0;

	taglist = msg->ops_AttrList;

	lplace   = GetTagData(GTSL_LevelPlace, PLACETEXT_LEFT, taglist);
	if (lplace & (PLACETEXT_ABOVE | PLACETEXT_BELOW))
	{
		/*
		 * Set correct Justification..the gadtools docs are wrong here
		 */
		justification = GTJ_CENTER;
	}

	/* Parse tags */
	while ((tag = NextTagItem((struct TagItem **)&taglist)))
	{
		IPTR tidata = tag->ti_Data;

		switch (tag->ti_Tag)
		{
		case GA_Width:
			width = (int)tidata;
			break;
		case GA_Height:
			height = (int)tidata;
			break;

		case GA_RelVerify:
			relverify = tidata != 0;
			break;
		case GA_Immediate:
			immediate = tidata != 0;
			break;

		case GA_DrawInfo:
			stags[4].ti_Data = tidata;
			dri = (APTR)tidata;
			break;

		case GA_TextAttr:
			tattr = (struct TextAttr *)tidata;
			break;

		/* 31-Mar-2003 bugfix
		 * Relay GA_UserData to super - Piru
		 */
		case GA_UserData:
			stags[5].ti_Data = tidata;
			break;

		case GTSL_Min:
			stags[6].ti_Data = tidata;
			break;
		case GTSL_Max:
			stags[7].ti_Data = tidata;
			break;
		case GTSL_Level:
			level = stags[8].ti_Data = tidata;
			break;

		case PGA_Freedom:
			if (tidata == LORIENT_HORIZ)
				stags[9].ti_Data = FREEHORIZ;
			else
				stags[9].ti_Data = FREEVERT;
			break;

		case GTSL_LevelFormat:
			lformat       = (STRPTR)tidata;
			break;
		case GTSL_MaxLevelLen:
			lmaxlen       = (UWORD)tidata;
			break;
		case GTSL_MaxPixelLen:
			lmaxpixellen  = (ULONG)tidata;
			break;
			//	    case GTSL_LevelPlace:    lplace        = (UBYTE)tidata;	break;
		case GTSL_Justification:
			justification = tidata;
			break;
		case GTSL_DispFunc:
			dispfunc      = (APTR)tidata;
			break;
		}

	} /* while (iterate taglist) */

	if (!dri)
		return NULL;

	stags[2].ti_Data = width - BORDERPROPSPACINGX * 2;
	stags[3].ti_Data = height - BORDERPROPSPACINGY * 2;

	/* Create slider gadget */
	slidergad = NewObjectA(GadToolsBase->sliderclass, NULL, stags);

	DEBUG_CREATESLIDER(bug("sliderg::New: slidergad 0x%lx\n",slidergad));

	if (!slidergad)
		return (NULL);

	sgtags[1].ti_Data = (IPTR)msg->ops_AttrList;

	newmsg.MethodID = OM_NEW;
	newmsg.ops_AttrList = sgtags;
	newmsg.ops_GInfo = NULL;

	o = (Object *)DoSuperMethodA(cl, o, (Msg)&newmsg);
	if (o)
	{
		struct SGData *data = INST_DATA(cl, o);
		struct opMember addm;
		struct TagItem tags[3];

		DEBUG_CREATESLIDER(bug("sliderg::New: sliderg %p\n", o));

		data->slider = slidergad;
		data->immediate = immediate;
		data->relverify = relverify;
		data->dispfunc = dispfunc;
		data->justification = justification;

		if (tattr)
			data->font = OpenFont(tattr);

		if (lformat && lmaxlen)
		{
			UWORD xsize;

			ysize = dri->dri_Font->tf_YSize;
			xsize = dri->dri_Font->tf_XSize;

			if (data->font)
			{
				ysize = data->font->tf_YSize;
				xsize = data->font->tf_XSize;
			}

			DEBUG_CREATESLIDER(bug("sliderg::New: font xsize %ld ysize %ld\n",xsize,ysize));

			if (!lmaxpixellen)
			{
				lmaxpixellen = lmaxlen * xsize;

				DEBUG_CREATESLIDER(bug("sliderg::New: lmaxpixellen %ld\n",lmaxpixellen));

			} /* if (!lmaxpixellen) */

			switch (lplace)
			{
			default:
			case PLACETEXT_LEFT:
				x = -lmaxpixellen - 4;
				y = (height - ysize) / 2 + 1;
				break;
			case PLACETEXT_RIGHT:
				x = width + 5;
				y = (height - ysize) / 2 + 1;
				break;
			case PLACETEXT_ABOVE:
				x = (width - lmaxpixellen) / 2;
				y = -ysize - 2;
				break;
			case PLACETEXT_BELOW:
				x = (width - lmaxpixellen) / 2;
				y = height + 3;
				break;
			}

			data->level_x = x;
			data->level_y = y;
			data->level_width = lmaxpixellen;
			data->level_height = ysize;
		}
		else
			lformat = NULL;

		data->levelformat = lformat;

		addm.MethodID = OM_ADDMEMBER;
		addm.opam_Object = slidergad;
		DoMethodA(o, (Msg)&addm);

		tags[0].ti_Tag = ICA_TARGET;
		tags[0].ti_Data = (IPTR)o;
		tags[1].ti_Tag = TAG_END;

		SetAttrsA(slidergad, tags);

		tags[0].ti_Tag = GA_Width;
		tags[0].ti_Data = width;
		tags[1].ti_Tag = GA_Height;
		tags[1].ti_Data = height;
		tags[2].ti_Tag = TAG_END;

		SetAttrsA(o, tags);

		ReturnPtr ("slider::New", IPTR, (IPTR)o);
	}

	ReturnPtr ("slider::New", IPTR, 0);
}

/**********************************************************************************************/

STATIC IPTR sliderg_dispose(Class * cl, Object * o, Msg msg)
{
	struct SGData *data = INST_DATA(cl, o);
	if (data->font) CloseFont(data->font);
	return DoSuperMethodA(cl, o, msg);
}

/**********************************************************************************************/

STATIC IPTR sliderg_set(Class * cl, Object * o, struct opUpdate * msg)
{
	IPTR		retval = 0UL;
	struct SGData	*data = INST_DATA(cl, o);
	struct TagItem	*tag, *tstate;
	struct TagItem	stags[] =
	    {
		    {
			    TAG_IGNORE,	0
		    },
		    {TAG_IGNORE,	0	},
		    {TAG_IGNORE,	0	},
		    {TAG_DONE		}
	    };
	BOOL has_stag = FALSE, has_non_stag = FALSE;

	D(bug("sliderg_set(cl 0x%lx o 0x%lx msg 0x%lx)\n",cl,o,msg));

	tstate = msg->opu_AttrList;

	while ((tag = NextTagItem(&tstate)))
	{
		D(bug("sliderg_set: tag 0x%08lx 0x%08lx\n", tag->ti_Tag, tag->ti_Data));

		switch (tag->ti_Tag)
		{
		case GTSL_Min:
			stags[0].ti_Tag = GTSL_Min;
			stags[0].ti_Data = tag->ti_Data;
			has_stag = TRUE;
			break;
		case GTSL_Max:
			stags[1].ti_Tag = GTSL_Max;
			stags[1].ti_Data = tag->ti_Data;
			has_stag = TRUE;
			break;
		case GTSL_Level:
			stags[2].ti_Tag = GTSL_Level;
			stags[2].ti_Data = tag->ti_Data;
			has_stag = TRUE;
			if ((WORD)tag->ti_Data != data->last_level)
			{
				data->draw_level = TRUE;
				retval = 1;
			}
			break;

		case GTSL_LevelFormat:
			data->levelformat = (APTR)tag->ti_Data;
			data->draw_level = TRUE;
			retval = 1;
			break;
		case GTSL_DispFunc:
			data->dispfunc = (APTR)tag->ti_Data;
			data->draw_level = TRUE;
			retval = 1;
			break;
		case GTSL_Justification:
			data->justification = (UBYTE)tag->ti_Data;
			data->draw_level = TRUE;
			retval = 1;
			break;

		default:
			has_non_stag = TRUE;
			break;
		}

	}

	if (has_stag)
	{
		struct opUpdate method;
		struct SGData *data = INST_DATA(cl, o);

		D(bug("sliderg_set: set slider tags\n"));

		method.MethodID = msg->MethodID;
		method.opu_AttrList = stags;
		method.opu_GInfo = msg->opu_GInfo;
		method.opu_Flags = msg->opu_Flags; /* only valid for OM_UPDATE. Oh well. */

		retval += DoMethodA(data->slider, (Msg)&method);
	}

	if (has_non_stag)
	{
		D(bug("sliderg_set: set superclass tags\n"));

		retval += DoSuperMethodA(cl, o, (Msg)msg);
	}

	if ((retval) && (OCLASS(o) == cl))
	{
		struct RastPort *rp = ObtainGIRPort(msg->opu_GInfo);
		if (rp)
		{
			DoMethod(o, GM_RENDER, (ULONG) msg->opu_GInfo, (ULONG) rp, GREDRAW_UPDATE);
			ReleaseGIRPort(rp);
			retval = FALSE;
		}
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR sliderg_get(Class * cl, Object * o, struct opGet *msg)
{
	IPTR		retval;

	switch(msg->opg_AttrID)
	{
	case GTA_GadgetKind:
	case GTA_ChildGadgetKind:
		*msg->opg_Storage = SLIDER_KIND;
		retval = 1UL;
		break;

	case GTSL_Min:
	case GTSL_Max:
	case GTSL_Level:
		{
			struct SGData *data = INST_DATA(cl, o);
			retval = DoMethodA(data->slider, (Msg)msg);
		}
		break;

		//case GA_Disabled:
	default:
		retval = DoSuperMethodA(cl, o, (Msg)msg);
		break;
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR sliderg_render(Class *cl, Object *o, struct gpRender *msg)
{
	IPTR		retval;
	struct SGData	*data = INST_DATA(cl, o);

	D(bug("sliderg::Render: draw_level %ld\n"));

	if ((msg->gpr_Redraw == GREDRAW_REDRAW || data->draw_level) && data->levelformat)
	{
		struct RastPort *rp = msg->gpr_RPort;
		struct TextFont *oldfont = NULL;       /* shut up the compiler */
		WORD x, y, w, h;
		LONG val;
		UBYTE buf[256];
		UBYTE *str = buf;
		ULONG len;
		struct TextExtent te;

		if (data->font)
		{
			oldfont = rp->Font;
			SetFont(rp, data->font);
		}

		if (data->erase_x1 <= data->erase_x2)
		{
			EraseRect(rp, data->erase_x1, data->erase_y1, data->erase_x2, data->erase_y2);
		}

		GetAttr(GTSL_Level, data->slider, &val);

		data->last_level = (WORD)val;

		D(bug("sliderg::Render: val %ld data->dispfunc 0x%lx\n", val,data->dispfunc));

		if (data->dispfunc)
		{
#ifdef __MORPHOS__
			REG_A7 -= 8;
			((ULONG *)REG_A7)[0] = (ULONG)data->slider;
			((ULONG *)REG_A7)[1] = (UWORD)val;
						REG_A1 = (ULONG)data->slider;   // some apps rely on register
						REG_D0 = (UWORD)val;            // parameters (cyfm 22/02/03)
			val = MyEmulHandle->EmulCallDirect68k(data->dispfunc);
			REG_A7 += 8;
#else
			val = ((LONG(*)(struct Gadget *, WORD))data->dispfunc)((struct Gadget *)data->slider, (WORD)val);
#endif
				D(bug("sliderg::Render: val -> val %ld\n", val));
		}

		RawDoFmt(data->levelformat, &val, (VOID_FUNC)AROS_ASMSYMNAME(puttostr), &str);

		len = strlen(buf);

		TextExtent(rp, buf, len, &te);

		x = G(o)->LeftEdge + data->level_x;
		y = G(o)->TopEdge + data->level_y;
		w = data->level_width;
		h = data->level_height;

		x -= te.te_Extent.MinX;
		y -= te.te_Extent.MinY;

		switch (data->justification)
		{
		case GTJ_RIGHT:
			x += w - te.te_Width;
			break;

		case GTJ_CENTER:
			x += (w - te.te_Width) / 2;
			break;

		case GTJ_LEFT:
			break;
		}

		data->erase_x1 = x + te.te_Extent.MinX;
		data->erase_y1 = y + te.te_Extent.MinY;
		data->erase_x2 = x + te.te_Extent.MaxX;
		data->erase_y2 = y + te.te_Extent.MaxY;

		SetABPenDrMd(rp, 1, 0, JAM1);
		Move(rp, x, y);
		Text(rp, buf, len);

		if (data->font)
			SetFont(rp, oldfont);

		data->draw_level = FALSE;
	}

	retval = DoSuperMethodA(cl, o, (Msg)msg);

	ReturnInt("slider::Render", IPTR, retval);
}

/**********************************************************************************************/

STATIC IPTR sliderg_gadgetidcmpmsg(Class *cl, Object *o, struct gtpGadgetIDCMPMsg *msg)
{
	struct SGData *data = INST_DATA(cl, o);
	IPTR val;

	switch (msg->IntuiMsg->Class)
	{
	case IDCMP_GADGETDOWN:
		GetAttr(GTSL_Level, data->slider, &val);
		if (!data->immediate)
		{
			if (data->code == (UWORD)val)
				return TRUE;
			msg->IntuiMsg->Class = IDCMP_MOUSEMOVE;
		}
		msg->IntuiMsg->Code = data->code = (UWORD)val;
		break;

	case IDCMP_MOUSEMOVE:
	//case IDCMP_INTUITICKS:
		GetAttr(GTSL_Level, data->slider, &val);
		if (data->code == (UWORD)val)
			return TRUE;
		//msg->IntuiMsg->Class = IDCMP_MOUSEMOVE;
		msg->IntuiMsg->Code = data->code = (UWORD)val;
		break;

	case IDCMP_GADGETUP:
		GetAttr(GTSL_Level, data->slider, &val);
		if (!data->relverify)
		{
			if (data->code == (UWORD)val)
				return TRUE;
			msg->IntuiMsg->Class = IDCMP_MOUSEMOVE;
		}
		msg->IntuiMsg->Code = data->code = (UWORD)val;
		break;
	}

	return FALSE;
}

/**********************************************************************************************/

AROS_UFH3S(IPTR, dispatch_slidergclass,
		   AROS_UFHA(Class *,  cl,  A0),
		   AROS_UFHA(Object *, o,   A2),
		   AROS_UFHA(Msg,      msg, A1)
		  )
{
	AROS_USERFUNC_INIT

	IPTR retval;

	D(bug("dispatch_slidergclass: cl 0x%lx o 0x%lx msg 0x%lx\n",cl,o,msg));

	switch(msg->MethodID)
	{
	case OM_NEW:
		D(bug("dispatch_slidergclass: OM_NEW\n"));
		retval = sliderg_new(cl, o, (struct opSet *)msg);
		break;

	case OM_DISPOSE:
		D(bug("dispatch_slidergclass: OM_DISPOSE\n"));
		retval = sliderg_dispose(cl, o, msg);
		break;

	case OM_UPDATE:
	case OM_SET:
		D(bug("dispatch_slidergclass: %s\n",
		      msg->MethodID == OM_SET ? "OM_SET" : "OM_UPDATE"));
		retval = (IPTR)sliderg_set(cl, o, (struct opUpdate *)msg);

		/* If we have been subclassed, OM_UPDATE should not cause a GM_RENDER
		 * because it would circumvent the subclass from fully overriding it.
		 * The check of cl == OCLASS(o) should fail if we have been
		 * subclassed, and we have gotten here via DoSuperMethodA().
		 */
		if (retval && (msg->MethodID == OM_UPDATE) && (cl == OCLASS(o)))
		{
			struct GadgetInfo *gi = ((struct opSet *)msg)->ops_GInfo;
			if (gi)
			{
				struct RastPort *rp = ObtainGIRPort(gi);
				if (rp)
				{
					DoMethod(o, GM_RENDER, (ULONG) gi, (ULONG) rp, GREDRAW_REDRAW);
					ReleaseGIRPort(rp);
				} /* if */
			} /* if */
		} /* if */
		break;

	case OM_GET:
		D(bug("dispatch_slidergclass: OM_GET\n"));
		retval = (IPTR)sliderg_get(cl, o, (struct opGet *)msg);
		break;

	case GM_RENDER:
		D(bug("dispatch_slidergclass: OM_RENDER\n"));
		retval = sliderg_render(cl, o, (struct gpRender *)msg);
		break;

	case GTM_GadgetIDCMPMsg:
		D(bug("dispatch_slidergclass: GTM_GadgetIDCMPMsg\n"));
		retval = sliderg_gadgetidcmpmsg(cl, o, (struct gtpGadgetIDCMPMsg *)msg);
		break;

	default:
		retval = DoSuperMethodA(cl, o, msg);
		break;
	} /* switch */

	D(bug("dispatch_slidergclass: retval 0x%lx\n",retval));

	return (retval);

	AROS_USERFUNC_EXIT
}  /* dispatch_Gtlvclass */

/**********************************************************************************************/

#undef GadToolsBase

Class *makeslidergclass(struct GadToolsBase_intern * GadToolsBase)
{
	Class *cl;

	ObtainSemaphore(&GadToolsBase->classsema);

	cl = GadToolsBase->slidergclass;
	if (!cl)
	{
		if (makegroupclass(GadToolsBase) && makesliderclass(GadToolsBase))
		{
			cl = MakeClass(NULL, NULL, GadToolsBase->groupclass, sizeof(struct SGData), 0UL);
			if (cl)
			{
				cl->cl_Dispatcher.h_Entry = (APTR) AROS_ASMSYMNAME(dispatch_slidergclass);
				cl->cl_Dispatcher.h_SubEntry = NULL;
				cl->cl_UserData = (IPTR) GadToolsBase;

				GadToolsBase->slidergclass = cl;
			}
		}
	}

	ReleaseSemaphore(&GadToolsBase->classsema);

	return cl;
}

/**********************************************************************************************/

