/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: stringclass.c,v 1.2 2006/08/21 21:06:28 piru Exp $
 
    Desc: Internal GadTools string class.
    Lang: English
*/

#undef AROS_ALMOST_COMPATIBLE
#define AROS_ALMOST_COMPATIBLE 1

#include <proto/exec.h>
#include <exec/libraries.h>
#include <exec/memory.h>
#include <proto/dos.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/intuition.h>
#include <intuition/cghooks.h>
#include <graphics/rastport.h>
#include <graphics/text.h>
#include <utility/tagitem.h>
#include <devices/inputevent.h>
#include <proto/alib.h>
#include <proto/utility.h>

#include <string.h> /* memset() */

#if 0
#define SDEBUG 1
#define DEBUG 1
#endif

#include "gadtools_intern.h"

/**********************************************************************************************/

#define G(x) ((struct Gadget *)(x))
#define EG(x) ((struct ExtGadget *)(x))

#define GadToolsBase ((struct GadToolsBase_intern *)cl->cl_UserData)

/**********************************************************************************************/

struct StringData
{
	Object		*frame;
	struct TextFont	*font;
	UBYTE		labelplace;
};

/**********************************************************************************************/

STATIC IPTR string_new(Class *cl, Object *o, struct opSet *msg)
{
	struct TagItem *tag, *taglist;
	struct opSet method;
	struct DrawInfo	*dri = NULL;
	struct TextAttr	*tattr = NULL;
	struct IBox		bbox;
	ULONG		labelplace = GV_LabelPlace_Left;

	struct TagItem tags[] =
	    {
		    {
			    GA_Left		, BORDERSTRINGSPACINGX
		    }
		    , /* 0 */
		    {GA_Top			, BORDERSTRINGSPACINGY	}, /* 1 */
		    {GA_Width		, 0			}, /* 2 */
		    {GA_Height		, 0			}, /* 3 */
		    {GA_TabCycle		, TRUE			}, /* 4 */
		    {STRINGA_TextVal	, (IPTR)NULL		}, /* 5 */
		    {STRINGA_MaxChars	, 65UL			}, /* 6 */
		    {STRINGA_EditHook	, (IPTR)NULL		}, /* 7 */
		    {STRINGA_ExitHelp	, FALSE			}, /* 8 */
		    {STRINGA_Justification	, GACT_STRINGLEFT	}, /* 9 */
		    {STRINGA_ReplaceMode	, FALSE			}, /* 10 */
		    {GA_TextAttr		, (IPTR)NULL		}, /* 11 */
		    {GA_UserData		, (IPTR)NULL		}, /* 12 */
		    {GA_ID			, 0			}, /* 13 */
		    {GA_Bounds		, (IPTR)&bbox		}, /* 14 */
		    {GA_Previous		, (IPTR)NULL		}, /* 15 */
		    {GA_Immediate		, FALSE			}, /* 16 */
		    {GA_RelVerify		, TRUE			}, /* 17 */
		    {GA_IntuiText		, (IPTR)NULL		}, /* 18 */
		    {TAG_DONE					}  /* 19 */
	    };


	DEBUG_CREATESTRING(bug("makestring:\n"));

	/* Could use GetTagData(), but this is faster */
	taglist = msg->ops_AttrList;
	while ((tag = NextTagItem((struct TagItem **)&taglist)))
	{
		IPTR tidata = tag->ti_Data;

		switch (tag->ti_Tag)
		{
		case GA_Left:
			tags[0].ti_Data = tidata + BORDERSTRINGSPACINGX;
			bbox.Left = tidata;
			break;
		case GA_Top:
			tags[1].ti_Data = tidata + BORDERSTRINGSPACINGY;
			bbox.Top = tidata;
			break;
		case GA_Width:
			tags[2].ti_Data = tidata - 2*BORDERSTRINGSPACINGX;
			bbox.Width = tidata;
			break;
		case GA_Height:
			tags[3].ti_Data = tidata - 2*BORDERSTRINGSPACINGY;
			bbox.Height = tidata;
			break;
		case GA_TabCycle:
			tags[4].ti_Data	= tidata;
			break;
		case GTIN_Number:
			tags[5].ti_Tag = STRINGA_LongVal;
			tags[5].ti_Data = tidata;
			break;
			/* Another weird inconsistency of AmigaOS GUI objects:
			** For intuition and strgclass gadgets, MaxChars includes trailing
			** zero, but this is NOT true for gadtools string gadgets
			*/
		case GTIN_MaxChars:
		case GTST_MaxChars:
			tags[6].ti_Data = tidata + 1;
			break;
			//case GTIN_EditHook:
		case GTST_String:
			/* Yet another inconsistency: gadtools handles GTST_String,
			** NULL, but intuition strgclass STRINGA_TextVal doesn't. - Piru
			*/
			tags[5].ti_Data	= tidata ? tidata : (ULONG)"";
			break;
		case GTST_EditHook:
			tags[7].ti_Data	= tidata;
			break;
		case STRINGA_ExitHelp:
			tags[8].ti_Data	= tidata;
			break;
		case STRINGA_Justification:
			tags[9].ti_Data	= tidata;
			break;
		case STRINGA_ReplaceMode:
			tags[10].ti_Data= tidata;
			break;
		case GA_TextAttr:
			tags[11].ti_Data= tidata;
			tattr = (APTR)tidata;
			break;
		case GA_UserData:
			tags[12].ti_Data= tidata;
			break;
		case GA_ID:
			tags[13].ti_Data= tidata;
			break;
		case GA_DrawInfo:
			dri = (struct DrawInfo *)tidata;
			break;
		case GA_Previous:
			tags[15].ti_Data= tidata;
			break;
		case GA_Immediate:
			tags[16].ti_Data= tidata;
			break;
		case GA_RelVerify:
			tags[17].ti_Data= tidata;
			break;
		case GA_LabelPlace:
			labelplace = tidata;
			break;
		case GA_IntuiText:
			tags[18].ti_Data= tidata;
			break;
		}

	} /* while (iterate taglist) */

	DEBUG_CREATESTRING(bug("makestring: tabcycle 0x%lx\n",
	                       tags[4].ti_Data));

	method.MethodID = OM_NEW;
	method.ops_AttrList = tags;
	method.ops_GInfo = NULL;

	o = (Object *)DoSuperMethodA(cl, o, (Msg)&method);

	if (o)
	{
		struct StringData *data = INST_DATA(cl, o);
		struct TagItem fitags[] =
		    {
			    {
				    IA_Width		, 0UL
			    },
			    {IA_Height		, 0UL		},
			    {IA_Resolution	, 0UL		},
			    {IA_FrameType	, FRAME_RIDGE	},
			    {IA_EdgesOnly	, TRUE		},
			    {TAG_DONE		, 0UL		}
		    };

		EG(o)->GadgetType |= GTYP_GADTOOLS;
		data->labelplace = labelplace;

		fitags[0].ti_Data = G(o)->Width;
		fitags[1].ti_Data = G(o)->Height;
		/* In theory dri can't be NULL here */
		if (dri)
			fitags[2].ti_Data = (dri->dri_Resolution.X << 16) + dri->dri_Resolution.Y;
		else
			fitags[2].ti_Tag = TAG_IGNORE;

		data->frame = NULL;

		if (tattr && o)
		{
			data->font = OpenFont(tattr);

			if (data->font)
			{
				struct TagItem sftags[] =
				    {
					    {
						    STRINGA_Font, (IPTR)NULL
					    }
					    , {TAG_DONE, }
				    };

				sftags[0].ti_Data = (IPTR)data->font;

				DoSuperMethod(cl, o, OM_SET, (ULONG) sftags, NULL);
			}
		}

		D(bug("Creating frame image"));

		data->frame = NewObjectA(NULL, FRAMEICLASS, fitags);

		D(bug("Created frame image: %p", data->frame));
		if (!data->frame)
		{
			CoerceMethod(cl, o, OM_DISPOSE);

			o = (IPTR)NULL;
		}

		ReturnPtr("string::New", IPTR, (IPTR)o);
	}

	ReturnPtr ("string::New", IPTR, 0);
}

/**********************************************************************************************/

STATIC IPTR string_set(Class *cl, Object *o, struct opSet *msg)
{
	struct TagItem	*tag, *tstate, tags[] =
	    {
		    {
			    TAG_IGNORE	, 0UL
		    }
		    ,  /* 0 STRINGA_TextVal  */
		    {TAG_MORE	, 0UL	}
	    };
	struct opSet	method;

	IPTR		retval = 0UL;

	EnterFunc(bug("String::SetNew()\n"));

	tstate = msg->ops_AttrList;
	while ((tag = NextTagItem((struct TagItem **)&tstate)))
	{
		IPTR tidata = tag->ti_Data;

		switch (tag->ti_Tag)
		{
		case GTST_String:
			/* Yet another inconsistency: gadtools handles GTST_String,
			** NULL, but intuition strgclass STRINGA_TextVal doesn't. - Piru
			*/
			tags[0].ti_Tag = STRINGA_TextVal;
			tags[0].ti_Data = tidata ? tidata : (ULONG)"";
			break;

		case GTIN_Number:
			tags[0].ti_Tag = STRINGA_LongVal;
			tags[0].ti_Data = tidata;
			break;
		}
	}

	tags[1].ti_Data = (IPTR)msg->ops_AttrList;

	method.MethodID = OM_SET;
	method.ops_AttrList = tags;
	method.ops_GInfo = msg->ops_GInfo;

	retval = DoSuperMethodA(cl, o, (Msg)&method);

#if 0
	if ((retval) && (OCLASS(o) == cl))
	{
		struct RastPort *rp = ObtainGIRPort(msg->ops_GInfo);
		if (rp)
		{
			DoMethod(o, GM_RENDER, (ULONG) msg->ops_GInfo, (ULONG) rp, GREDRAW_UPDATE);
			ReleaseGIRPort(rp);
			retval = FALSE;
		}
	}
#endif
	ReturnPtr ("String::SetNew", IPTR, retval);
}

/**********************************************************************************************/

STATIC IPTR string_get(Class *cl, Object *o, struct opGet *msg)
{
	struct opGet	cloned_msg = *msg;
	IPTR		retval = 0;

	switch(cloned_msg.opg_AttrID)
	{
	case GTA_GadgetKind:
	case GTA_ChildGadgetKind:
		*msg->opg_Storage = STRING_KIND;
		retval = 1;
		break;

	case GTST_String:
		cloned_msg.opg_AttrID = STRINGA_TextVal;
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;

	case GTIN_Number:
		cloned_msg.opg_AttrID = STRINGA_LongVal;
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;

	default:
		retval = DoSuperMethodA(cl, o, (Msg)&cloned_msg);
		break;
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR string_render(Class *cl, Object *o, struct gpRender *msg)
{
	IPTR retval;

	EnterFunc(bug("String::Render()\n"));
	retval = DoSuperMethodA(cl, o, (Msg)msg);

	D(bug("Superclass render OK\n"));

	if (msg->gpr_Redraw == GREDRAW_REDRAW)
	{
		struct StringData *data = INST_DATA(cl, o);
		struct RastPort *rp = msg->gpr_RPort;

		WORD x, y;

		struct TagItem itags[] =
		    {
			    {
				    IA_Width,	0L
			    },
			    {IA_Height,	0L},
			    {TAG_DONE,}
		    };

		D(bug("Full redraw\n"));

		/* center image position, we assume image top and left is 0 */
		itags[0].ti_Data = G(o)->Width + BORDERSTRINGSPACINGX * 2;
		itags[1].ti_Data = G(o)->Height + BORDERSTRINGSPACINGY * 2;

		SetAttrsA((Object *)data->frame, itags);

		x = G(o)->LeftEdge - BORDERSTRINGSPACINGX;
		y = G(o)->TopEdge - BORDERSTRINGSPACINGY;

		DrawImageState(rp,
		               (struct Image *)data->frame,
		               x, y,
		               IDS_NORMAL,
		               msg->gpr_GInfo->gi_DrInfo);

		renderlabel(GadToolsBase, (struct Gadget *)o, msg->gpr_RPort, data->font, data->labelplace);

	} /* if (whole gadget should be redrawn) */

	ReturnInt ("String::Render", IPTR, retval);
}

/**********************************************************************************************/

STATIC IPTR string_dispose(Class *cl, Object *o, Msg msg)
{
	struct StringData *data = INST_DATA(cl, o);

	if (data->frame) DisposeObject(data->frame);
	if (data->font) CloseFont(data->font);

	return DoSuperMethodA(cl, o, (Msg)msg);
}

/**********************************************************************************************/

STATIC IPTR string_gadgetidcmpmsg(Class *cl, Object *o, struct gtpGadgetIDCMPMsg *msg)
{
	D(bug("string_gadgetidcmpmsg: cl %p o %p msg %p\n",
	      cl,
	      o,
	      msg));

	D(bug("string_gadgetidcmpmsg: intuimsg %p class 0x%lx\n",
	      msg->IntuiMsg,
	      msg->IntuiMsg->Class));

	switch (msg->IntuiMsg->Class)
	{
	case IDCMP_GADGETDOWN:
		D(bug("string_gadgetidcmpmsg: IDCMP_GADGETDOWN\n"));
		return (EG(o)->Activation & GACT_IMMEDIATE) == 0;

	case IDCMP_MOUSEMOVE:
		D(bug("string_gadgetidcmpmsg: IDCMP_MOUSEMOVE\n"));
	case IDCMP_INTUITICKS:
		D(bug("string_gadgetidcmpmsg: IDCMP_INTUITICKS\n"));
		return TRUE;

	case IDCMP_GADGETUP:
		D(bug("string_gadgetidcmpmsg: IDCMP_GADGETUP\n"));
		break;
	}

	return FALSE;
}

/**********************************************************************************************/

AROS_UFH3S(IPTR, dispatch_stringclass,
           AROS_UFHA(Class *, cl, A0),
           AROS_UFHA(Object *, o, A2),
           AROS_UFHA(Msg, msg, A1)
          )
{
	AROS_USERFUNC_INIT

	IPTR retval;

	D(bug("dispatch_stringclass: %p o %p msg %p\n",cl,o,msg));

	switch (msg->MethodID)
	{
	case OM_NEW:
		D(bug("dispatch_stringclass: OM_NEW\n"));
		retval = string_new(cl, o, (struct opSet *)msg);
		break;

	case OM_UPDATE:
	case OM_SET:
		D(bug("dispatch_stringclass: %s\n",
		      msg->MethodID == OM_SET ? "OM_SET" : "OM_UPDATE"));
		retval = (IPTR)string_set(cl, o, (struct opSet *)msg);

		/* If we have been subclassed, OM_UPDATE should not cause a GM_RENDER
		 * because it would circumvent the subclass from fully overriding it.
		 * The check of cl == OCLASS(o) should fail if we have been
		 * subclassed, and we have gotten here via DoSuperMethodA().
		 */
		if (retval && (msg->MethodID == OM_UPDATE) && (cl == OCLASS(o)))
		{
			struct GadgetInfo *gi = ((struct opSet *)msg)->ops_GInfo;
			if (gi)
			{
				struct RastPort *rp = ObtainGIRPort(gi);
				if (rp)
				{
					DoMethod(o, GM_RENDER, (ULONG) gi, (ULONG) rp, GREDRAW_REDRAW);
					ReleaseGIRPort(rp);
				} /* if */
			} /* if */
		} /* if */
		break;

	case OM_GET:
		D(bug("dispatch_stringclass: OM_GET\n"));
		retval = string_get(cl, o, (struct opGet *)msg);
		break;

	case GM_RENDER:
		D(bug("dispatch_stringclass: OM_RENDER\n"));
		retval = string_render(cl, o, (struct gpRender *)msg);
		break;

	case OM_DISPOSE:
		D(bug("dispatch_stringclass: OM_DISPOSE\n"));
		retval = string_dispose(cl, o, msg);
		break;

	case GTM_GadgetIDCMPMsg:
		D(bug("dispatch_stringclass: GTM_GadgetIDCMPMsg\n"));
		retval = string_gadgetidcmpmsg(cl, o, (struct gtpGadgetIDCMPMsg *)msg);
		break;

	default:
		D(bug("dispatch_stringclass: gate MethodID %ld\n",msg->MethodID));
		retval = DoSuperMethodA(cl, o, msg);
		break;
	}
	D(bug("dispatch_stringclass: retval 0x%lx\n",retval));

	return retval;

	AROS_USERFUNC_EXIT
}

/**********************************************************************************************/

#undef GadToolsBase

Class *makestringclass(struct GadToolsBase_intern * GadToolsBase)
{
	Class *cl;

	ObtainSemaphore(&GadToolsBase->classsema);

	cl = GadToolsBase->stringclass;
	if (!cl)
	{
		cl = MakeClass(NULL, STRGCLASS, NULL, sizeof(struct StringData), 0UL);
		if (cl)
		{
			cl->cl_Dispatcher.h_Entry = (APTR) AROS_ASMSYMNAME(dispatch_stringclass);
			cl->cl_Dispatcher.h_SubEntry = NULL;
			cl->cl_UserData = (IPTR) GadToolsBase;

			GadToolsBase->stringclass = cl;
		}
	}

	ReleaseSemaphore(&GadToolsBase->classsema);

	return cl;
}

/**********************************************************************************************/

