/*
	Copyright  1995-2001, The AROS Development Team. All rights reserved.
	$Id: stringgclass.c,v 1.1.1.1 2005/03/15 15:54:23 laire Exp $
 
	Desc: Internal GadTools slider class.
	Lang: English
*/

#undef AROS_ALMOST_COMPATIBLE
#define AROS_ALMOST_COMPATIBLE 1

#include <proto/exec.h>
#include <exec/libraries.h>
#include <exec/memory.h>
#include <proto/dos.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/icclass.h>
#include <intuition/intuition.h>
#include <utility/tagitem.h>
#include <proto/alib.h>
#include <proto/utility.h>
#include <proto/gadtools.h>

#include <string.h> /* memset() */

#if 0
#define SDEBUG 1
#define DEBUG 1
#endif

#include "gadtools_intern.h"

/**********************************************************************************************/

#define G(x) ((struct Gadget *)(x))
#define EG(x) ((struct ExtGadget *)(x))

#define GadToolsBase ((struct GadToolsBase_intern *)cl->cl_UserData)

/**********************************************************************************************/

struct SGData
{
	Object		*string;
};

/**********************************************************************************************/

STATIC IPTR stringg_new(Class *cl, Object *o, struct opSet *msg)
{
	struct TagItem *tag, *taglist;
	Object *string;

	struct TagItem sgtags[] =
	{
		{GA_TabCycle,	 TRUE},
		{TAG_MORE,	    }
	};

	struct TagItem tags[] =
	{
	    {GA_Left	 , BORDERSTRINGSPACINGX}, /* 0 */
	    {GA_Top			, BORDERSTRINGSPACINGY	}, /* 1 */
	    {GA_Width		, 0			}, /* 2 */
	    {GA_Height		, 0			}, /* 3 */
	    {GA_TabCycle		, TRUE			}, /* 4 */
	    {GTST_String		, (IPTR)NULL		}, /* 5 */
	    {GTST_MaxChars		, 64UL			}, /* 6 */
	    {GTST_EditHook		, (IPTR)NULL		}, /* 7 */
	    {STRINGA_ExitHelp	, FALSE			}, /* 8 */
	    {STRINGA_Justification	, GACT_STRINGLEFT	}, /* 9 */
	    {STRINGA_ReplaceMode	, FALSE			}, /* 10 */
	    {GA_TextAttr		, (IPTR)NULL		}, /* 11 */
	    {GA_UserData		, (IPTR)NULL		}, /* 12 */
	    {GA_ID			, 0			}, /* 13 */
	    {TAG_DONE					}  /* 14 */
	};

	DEBUG_CREATESTRING(bug("makestring:\n"));

	/* Could use GetTagData(), but this is faster */
	taglist = msg->ops_AttrList;
	while ((tag = NextTagItem((struct TagItem **)&taglist)))
	{
		IPTR tidata = tag->ti_Data;

		switch (tag->ti_Tag)
		{
		case GA_Width:
			tags[2].ti_Data = tidata - 2*BORDERSTRINGSPACINGX;
			break;
		case GA_Height:
			tags[3].ti_Data = tidata - 2*BORDERSTRINGSPACINGY;
			break;
		case GA_TabCycle:
			tags[4].ti_Data	= tidata;
			break;
		case GTIN_Number:
			tags[5] = *tag;
			break;
		case GTIN_MaxChars:
			tags[6] = *tag;
			break;
			//case GTIN_EditHook:
		case GTST_String:
			tags[5].ti_Data	= tidata;
			break;
		case GTST_MaxChars:
			tags[6].ti_Data	= tidata;
			break;
		case GTST_EditHook:
			tags[7].ti_Data	= tidata;
			break;
		case STRINGA_ExitHelp:
			tags[8].ti_Data	= tidata;
			break;
		case STRINGA_Justification:
			tags[9].ti_Data	= tidata;
			break;
		case STRINGA_ReplaceMode:
			tags[10].ti_Data= tidata;
			break;
		case GA_TextAttr:
			tags[11].ti_Data= tidata;
			break;
		case GA_UserData:
			tags[12].ti_Data= tidata;
			break;
		case GA_ID:
			tags[13].ti_Data= tidata;
			break;
		}

	} /* while (iterate taglist) */

	DEBUG_CREATESTRING(bug("makestring: tabcycle 0x%lx\n",
	                       tags[4].ti_Data));

	string = NewObjectA(GadToolsBase->stringclass, NULL, tags);
	if (!string)
		return NULL;

	sgtags[1].ti_Data = (IPTR)msg->ops_AttrList;
	msg->ops_AttrList = sgtags;
	o = (Object *)DoSuperMethodA(cl, o, (Msg)msg);
	msg->ops_AttrList = (APTR)sgtags[1].ti_Data;

	if (o)
	{
		struct SGData *data = INST_DATA(cl, o);
		struct opMember addm;

		DEBUG_CREATESTRING(bug("string::New: string %p\n", o));

		data->string = string;

		addm.MethodID = OM_ADDMEMBER;
		addm.opam_Object = string;
		DoMethodA(o, (Msg)&addm);

		EG(o)->SpecialInfo = EG(string)->SpecialInfo;   // copy SpecialInfo of string object member
		// to the "public" Gadget pointer because
		// libs like gtlayout.library access it (11-07-02 cyfm)
		EG(o)->Flags &= ~GFLG_TABCYCLE;
		EG(o)->Flags |= EG(string)->Flags & GFLG_TABCYCLE;
		EG(o)->Flags |= GFLG_STRINGEXTEND;

		ReturnPtr ("string::New", IPTR, (IPTR)o);
	}

	ReturnPtr ("string::New", IPTR, 0);
}

/**********************************************************************************************/

STATIC IPTR stringg_set(Class * cl, Object * o, struct opUpdate * msg)
{
	IPTR		retval = 0UL;
	//struct SGData	*data = INST_DATA(cl, o);
	struct TagItem	*tag, *tstate;
	struct TagItem	stags[] =
	{
		{TAG_IGNORE, 0},
		{TAG_DONE		}
	};
	BOOL has_stag = FALSE, has_non_stag = FALSE;

	D(bug("stringg_set(cl 0x%lx o 0x%lx msg 0x%lx)\n",cl,o,msg));

	tstate = msg->opu_AttrList;

	while ((tag = NextTagItem(&tstate)))
	{
		D(bug("stringg_set: tag 0x%08lx 0x%08lx\n", tag->ti_Tag, tag->ti_Data));

		switch (tag->ti_Tag)
		{
		case GTIN_Number:
			stags[0] = *tag;
			has_stag = TRUE;
			break;

		case GTST_String:
			stags[0] = *tag;
			has_stag = TRUE;
			break;

		default:
			has_non_stag = TRUE;
			break;
		}
	}

	if (has_stag)
	{
		struct opUpdate method;
		struct SGData *data = INST_DATA(cl, o);

		D(bug("stringg_set: set slider tags\n"));

		method.MethodID = msg->MethodID;
		method.opu_AttrList = stags;
		method.opu_GInfo = msg->opu_GInfo;
		method.opu_Flags = msg->opu_Flags; /* only valid for OM_UPDATE. Oh well. */

		retval += DoMethodA(data->string, (Msg)&method);
	}

	if (has_non_stag)
	{
		D(bug("stringg_set: set superclass tags\n"));

		retval += DoSuperMethodA(cl, o, (Msg)msg);
	}

	if ((retval) && (OCLASS(o) == cl))
	{
		struct RastPort *rp = ObtainGIRPort(msg->opu_GInfo);
		if (rp)
		{
			DoMethod(o, GM_RENDER, (ULONG) msg->opu_GInfo, (ULONG) rp, GREDRAW_UPDATE);
			ReleaseGIRPort(rp);
			retval = FALSE;
		}
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR stringg_get(Class * cl, Object * o, struct opGet *msg)
{
	IPTR		retval;

	switch(msg->opg_AttrID)
	{
	case GTA_GadgetKind:
	case GTA_ChildGadgetKind:
		*msg->opg_Storage = STRING_KIND;
		retval = 1;
		break;

	case GTIN_Number:
	case GTST_String:
		{
			struct SGData *data = INST_DATA(cl, o);
			retval = DoMethodA(data->string, (Msg)msg);
		}
		break;

		//case GA_Disabled:
	default:
		retval = DoSuperMethodA(cl, o, (Msg)msg);
		break;
	}

	return retval;
}

/**********************************************************************************************/

STATIC IPTR stringg_gadgetidcmpmsg(Class *cl, Object *o, struct gtpGadgetIDCMPMsg *msg)
{
	D(bug("stringg_gadgetidcmpmsg: cl 0x%lx o 0x%lx msg 0x%lx\n",
	      cl,
	      o,
	      msg));

	D(bug("stringg_gadgetidcmpmsg: intuimsg 0x%lx class 0x%lx\n",
	      msg->IntuiMsg,
	      msg->IntuiMsg->Class));

	switch (msg->IntuiMsg->Class)
	{
	case IDCMP_GADGETDOWN:
		D(bug("stringg_gadgetidcmpmsg: IDCMP_GADGETDOWN\n"));
		return (EG(o)->Activation & GACT_IMMEDIATE) == 0;

	case IDCMP_MOUSEMOVE:
		D(bug("stringg_gadgetidcmpmsg: IDCMP_MOUSEMOVE\n"));
	case IDCMP_INTUITICKS:
		D(bug("stringg_gadgetidcmpmsg: IDCMP_INTUITICKS\n"));
		return TRUE;

	case IDCMP_GADGETUP:
		D(bug("stringg_gadgetidcmpmsg: IDCMP_GADGETUP\n"));
		break;
	}

	return FALSE;
}

/**********************************************************************************************/

AROS_UFH3S(IPTR, dispatch_stringgclass,
		   AROS_UFHA(Class *,  cl,  A0),
		   AROS_UFHA(Object *, o,   A2),
		   AROS_UFHA(Msg,      msg, A1)
		  )
{
	AROS_USERFUNC_INIT

	IPTR retval;

	D(bug("dispatch_stringgclass: cl 0x%lx o 0x%lx msg 0x%lx methodID 0x%lx\n",cl,o,msg,msg->MethodID));

	switch(msg->MethodID)
	{
	case OM_NEW:
		D(bug("dispatch_stringgclass: OM_NEW\n"));
		retval = stringg_new(cl, o, (struct opSet *)msg);
		break;

	case OM_UPDATE:
	case OM_SET:
		D(bug("dispatch_stringgclass: %s\n",
		      msg->MethodID == OM_SET ? "OM_SET" : "OM_UPDATE"));
		retval = (IPTR)stringg_set(cl, o, (struct opUpdate *)msg);

		/* If we have been subclassed, OM_UPDATE should not cause a GM_RENDER
		 * because it would circumvent the subclass from fully overriding it.
		 * The check of cl == OCLASS(o) should fail if we have been
		 * subclassed, and we have gotten here via DoSuperMethodA().
		 */
		if (retval && (msg->MethodID == OM_UPDATE) && (cl == OCLASS(o)))
		{
			struct GadgetInfo *gi = ((struct opSet *)msg)->ops_GInfo;
			if (gi)
			{
				struct RastPort *rp = ObtainGIRPort(gi);
				if (rp)
				{
					DoMethod(o, GM_RENDER, (ULONG) gi, (ULONG) rp, GREDRAW_REDRAW);
					ReleaseGIRPort(rp);
				} /* if */
			} /* if */
		} /* if */
		break;

	case OM_GET:
		D(bug("dispatch_stringgclass: OM_GET\n"));
		retval = (IPTR)stringg_get(cl, o, (struct opGet *)msg);
		break;

	case GTM_GadgetIDCMPMsg:
		D(bug("dispatch_stringgclass: GTM_GadgetIDCMPMsg\n"));
		retval = stringg_gadgetidcmpmsg(cl, o, (struct gtpGadgetIDCMPMsg *)msg);
		break;

	default:
		retval = DoSuperMethodA(cl, o, msg);
		break;
	} /* switch */

	D(bug("dispatch_stringgclass: retval 0x%lx\n",retval));

	return (retval);

	AROS_USERFUNC_EXIT
}

/**********************************************************************************************/

#undef GadToolsBase

Class *makestringgclass(struct GadToolsBase_intern * GadToolsBase)
{
	Class *cl;

	ObtainSemaphore(&GadToolsBase->classsema);

	cl = GadToolsBase->stringgclass;
	if (!cl)
	{
		if (makegroupclass(GadToolsBase) && makestringclass(GadToolsBase))
		{
			cl = MakeClass(NULL, NULL, GadToolsBase->groupclass, sizeof(struct SGData), 0UL);
			if (cl)
			{
				cl->cl_Dispatcher.h_Entry = (APTR) AROS_ASMSYMNAME(dispatch_stringgclass);
				cl->cl_Dispatcher.h_SubEntry = NULL;
				cl->cl_UserData = (IPTR) GadToolsBase;

				GadToolsBase->stringgclass = cl;
			}
		}
	}

	ReleaseSemaphore(&GadToolsBase->classsema);

	return cl;
}

/**********************************************************************************************/
